from collections import OrderedDict

from kntgen.model.figma_models import *
from kntgen.str_helpers import *


class FWidgetType(str, Enum):
    PAGE = 'page'
    BACKGROUND = 'bg'
    CONTAINER = 'container'
    SIZED_BOX = 'sb'
    ICON = 'ic'
    TEXT = 'tv'
    TEXT_FIELD = 'tf'
    IMAGE = 'img'
    BUTTON = 'btn'
    SWITCHER = 'sw'
    CHECK_BOX = 'cb'
    COLUMN = 'column'
    ROW = 'row'
    WRAP = 'wrap'
    WRAP_VERTICAL = 'wrapv'
    STACK = 'stack'
    LIST_VIEW = 'lv'
    LIST_VIEW_HORIZONTAL = 'lvh'
    GRID_VIEW = 'gv'
    GRID_VIEW_HORIZONTAL = 'gvh'
    NEW_WIDGET = 'f'
    # Add built-in knt_ui
    TAB_PAGE = 'tpage'
    APP_BAR = 'ab'
    TOOL_BAR = 'tb'
    BOTTOM_BAR = 'bb'
    GALLERY = 'gallery'
    CALENDAR = 'calendar'
    DIVIDER = 'div'
    WEB = 'web'
    CHART = 'chart'
    IGNORED = 'ignored'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    def is_view_group(self):
        return self.is_flex_view_group() or self.is_consistent_view_group()

    def is_flex_view_group(self):
        """
        Flutter View Types which Flutter View is a view group with dynamic child
        style and size

        :return: Whether this type is a flex view group type or not
        """
        return self in [
            FWidgetType.COLUMN,
            FWidgetType.ROW,
            FWidgetType.WRAP,
            FWidgetType.STACK
        ]

    def is_consistent_view_group(self):
        """
        Flutter View Types which Flutter View is a view group with every child has
        the same structure

        :return: Whether this type is a consistent view group type or not
        """
        return self in [
            FWidgetType.LIST_VIEW,
            FWidgetType.LIST_VIEW_HORIZONTAL,
            FWidgetType.GRID_VIEW,
            FWidgetType.GRID_VIEW_HORIZONTAL
        ]


class Alignment(str, Enum):
    START = 'start'
    CENTER = 'center'
    END = 'end'
    SPACE_BETWEEN = 'spaceBetween'


class Direction(str, Enum):
    VERTICAL = 'vertical'
    HORIZONTAL = 'horizontal'


class Edge(str, Enum):
    LEFT = 'left'
    TOP = 'top'
    RIGHT = 'right'
    BOTTOM = 'bottom'

    @staticmethod
    def values():
        return [Edge.LEFT, Edge.TOP, Edge.RIGHT, Edge.BOTTOM]


class InsetType(str, Enum):
    PADDING = 'padding'
    MARGIN = 'margin'


class Gradient(str, Enum):
    LINEAR = 'GRADIENT_LINEAR'
    RADIAL = 'GRADIENT_RADIAL'
    SWEEP = 'GRADIENT_ANGULAR'

    def f_class_name(self):
        return {
            Gradient.RADIAL: 'RadialGradient',
            Gradient.SWEEP: 'SweepGradient'
        }.get(self, 'LinearGradient')

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @staticmethod
    def convert_coordinate_figma_to_flutter(coordinate: dict):
        if not coordinate or 'x' not in coordinate or 'y' not in coordinate:
            return 0.0, 0.0
        x = coordinate['x']
        y = coordinate['y']
        return round((x - 0.5) * 2, 2), round((y - 0.5) * 2, 2)


class FWidgetExtractor:
    REGEX_F_WIDGET = r'((?P<type>[A-Za-z0-9]+)_?)(?P<name>[A-Za-z0-9-_]+)?'
    FW_CLICKABLE = 'click_'
    FW_DYNAMIC_DATA = 'obj_'
    FW_CUSTOM_ITEM = 'item_'
    FW_CUSTOM_WIDGET = f'{FWidgetType.NEW_WIDGET.value}_'
    DEFAULT_COLORS = 'Colors.black'

    @staticmethod
    def figma_to_raw_flutter(current_node: dict):
        if 'name' not in current_node:
            return None
        # Defined view information
        *redundant, w_type = FWidgetExtractor.regex_info(current_node)
        widget = FRawWidget(figma_node=current_node)
        widget.type = w_type
        return widget

    @staticmethod
    def regex_info(current_node: dict):
        return FWidgetExtractor.regex_info_from_name(
            current_node['name'] if 'name' in current_node else '')

    @staticmethod
    def regex_info_from_name(v_name: str):
        w_full_name = v_name or ''
        w_name = ''
        w_type = FWidgetType.IGNORED
        w_full_name_for_matcher = w_full_name \
            .replace(FWidgetExtractor.FW_CLICKABLE, '') \
            .replace(FWidgetExtractor.FW_DYNAMIC_DATA, '')
        # In case this node is defined as a new widget
        if w_full_name_for_matcher.startswith(
                FWidgetExtractor.FW_CUSTOM_WIDGET):
            w_full_name_for_matcher = w_full_name_for_matcher.replace(
                FWidgetExtractor.FW_CUSTOM_WIDGET, '')
            w_type = FWidgetType.NEW_WIDGET
            w_name = w_full_name_for_matcher
            w_full_name = w_full_name_for_matcher
        # Continue to find info from regex
        match_view_info = re.search(
            FWidgetExtractor.REGEX_F_WIDGET,
            w_full_name_for_matcher
        )
        if match_view_info:
            match_type = match_view_info.group('type')
            match_name = match_view_info.group('name')
            converted_type = FWidgetType(match_type) \
                if FWidgetType.has_value(match_type) \
                else FWidgetType.IGNORED
            if converted_type != FWidgetType.IGNORED:
                w_type = converted_type
            if match_name:
                w_name = match_name.replace(FWidgetExtractor.FW_CUSTOM_ITEM, '') \
                    if match_name.startswith(FWidgetExtractor.FW_CUSTOM_ITEM) \
                    else match_name
        return w_full_name, w_name, w_type

    @staticmethod
    def name_without_type(v_name: str):
        _, text_without_type, _ = FWidgetExtractor.regex_info_from_name(v_name)
        return text_without_type

    @staticmethod
    def localized_key_from_name(v_name: str):
        """
        Convert Figma's node name into locale keys

        :param v_name: Name of a Figma node
        :return: A tuple contains (prefix such as page/item/custom, literal locale keys)
        """
        if not v_name:
            return '', ''
        name_without_type = FWidgetExtractor.name_without_type(v_name)
        prefix = ''
        if v_name.startswith(FWidgetExtractor.FW_CUSTOM_WIDGET):
            prefix = 'custom'
            v_name_remove_customized = v_name.replace(
                FWidgetExtractor.FW_CUSTOM_WIDGET, '')
            if v_name_remove_customized.startswith(f'{FWidgetType.PAGE}_'):
                prefix = 'page'
            if v_name_remove_customized.startswith(
                    FWidgetExtractor.FW_CUSTOM_ITEM):
                prefix = 'item'
        return prefix, name_without_type

    @staticmethod
    def text_related(current_node: dict):
        if not current_node:
            return None
        if not FWidgetExtractor.is_node_as_text_view(current_node):
            return None
        if 'type' not in current_node or current_node['type'] != 'TEXT':
            return None
        text_in_node = current_node[
            'characters'] if 'characters' in current_node else ''
        if not FWidgetExtractor.valid_content_in_text_node(text_in_node):
            return None
        # Flutter just supports only 1 text color with possible alpha mode
        hex_text_color = ''
        alpha_text_color = 255
        if 'fills' in current_node:
            if len(current_node['fills']) > 0:
                hex_text_color, alpha_text_color = FWidgetExtractor.get_hex_in_color_node(
                    current_node['fills'][0])
        return FTextRelated(
            character=text_in_node,
            character_style_override=current_node[
                'characterStyleOverrides'] if 'characterStyleOverrides'
                                              in current_node else [],
            hex_text_color=hex_text_color,
            alpha_text_color=alpha_text_color,
            style=TextStyle(**current_node[
                'style']) if 'style' in current_node else None,
            style_override_table=current_node[
                'styleOverrideTable'] if 'styleOverrideTable'
                                         in current_node else []
        )

    @staticmethod
    def color_related(current_node: dict) -> list[str]:
        def extract_color(component_color_node):
            hex_c_list = []
            for component in component_color_node:
                hex_c, _ = FWidgetExtractor.get_hex_in_color_node(component)
                if hex_c:
                    hex_c_list.append(hex_c)
                    continue
                # Dry-run to get gradient colors
                gradient_hex_c_list = FWidgetExtractor.gradient_color_related(
                    component)
                hex_c_list.extend([c[0] for c in gradient_hex_c_list if c])
            return hex_c_list

        hex_colors = []
        if 'fills' in current_node:
            hex_colors.extend(extract_color(current_node['fills']))
        if 'strokes' in current_node:
            hex_colors.extend(extract_color(current_node['strokes']))
        if 'effects' in current_node:
            hex_colors.extend(extract_color(current_node['effects']))
        return hex_colors

    @staticmethod
    def gradient_color_related(current_node: dict) -> list[tuple]:
        """
        Extract colors/alpha/stop positions/vector from gradient object

        :param current_node: Node contains gradient information
        :return: A tuple of (colors, alpha, stop positions, vector)
        """
        hex_c_list_with_position = []
        if 'type' in current_node \
                and current_node['type'].startswith('GRADIENT'):
            if 'gradientStops' in current_node \
                    and current_node['gradientStops']:
                # Get list vector presents gradient
                position_vector = current_node.get('gradientHandlePositions',
                                                   [])
                for gradient_color_node in current_node['gradientStops']:
                    hex_c, alpha_c = FWidgetExtractor.get_hex_in_color_node(
                        gradient_color_node)
                    position = gradient_color_node[
                        'position'] if 'position' in gradient_color_node else 0
                    if hex_c:
                        hex_c_list_with_position.append(
                            (hex_c, alpha_c, position, position_vector))
        return hex_c_list_with_position

    @staticmethod
    def get_hex_in_color_node(component):
        if 'color' in component and Color.valid_color_properties(
                component['color']):
            dict_c = component['color']
            alpha_c = math.floor(dict_c.get('a', 1) * 255)
            tuple_c = (
                math.floor(dict_c['r'] * 255),
                math.floor(dict_c['g'] * 255),
                math.floor(dict_c['b'] * 255)
            )
            hex_c = Color.rbg_to_hex(*tuple_c)
            if 'opacity' in component:
                alpha_c = math.floor(component.get('opacity', 1) * 255)
            return hex_c, alpha_c
        return '', 255

    @staticmethod
    def present_f_color(component,
                        colors_mapping: dict,
                        property_name: str = None):
        hex_color, alpha = FWidgetExtractor.get_hex_in_color_node(component)
        return FWidgetExtractor.present_f_color_from_scratch(
            hex_color=hex_color,
            alpha=alpha,
            colors_mapping=colors_mapping,
            property_name=property_name
        )

    @staticmethod
    def present_f_color_from_scratch(hex_color,
                                     alpha,
                                     colors_mapping: dict,
                                     property_name: str = None,
                                     need_prefix=True):
        named_color = FWidgetExtractor.hex_to_named_color(hex_color,
                                                          colors_mapping)
        property_present = property_name or 'color'
        prefix = f'{property_present}:' if need_prefix else ''
        return f'{prefix}{named_color}' \
               f'{FWidgetExtractor.present_f_color_alpha(alpha)}' \
            if named_color else ''

    @staticmethod
    def present_f_color_alpha(alpha):
        return f'.withAlpha({alpha})' if alpha is not None and alpha < 255 else ''

    @staticmethod
    def hex_to_named_color(hex_color: str, colors_mapping: dict):
        if not hex_color:
            return hex_color
        if hex_color in colors_mapping:
            hex_color = colors_mapping[hex_color] \
                        or FWidgetExtractor.DEFAULT_COLORS
        return hex_color

    @staticmethod
    def valid_content_in_text_node(content: str):
        if len(content) < 2:
            # Too short
            return False
        if is_number(content):
            # Number is meaningless when extract to mobile app strings
            return False
        if is_time_format(content):
            # Time format is meaningless
            return False
        return True

    @staticmethod
    def as_rect(figma_node):
        if 'absoluteBoundingBox' in figma_node:
            return RectLTRB(**figma_node['absoluteBoundingBox'])
        return RectLTRB(0, 0, 0, 0)

    @staticmethod
    def is_node_as_custom_widget(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return v_name.startswith(FWidgetExtractor.FW_CUSTOM_WIDGET)

    @staticmethod
    def is_node_as_page(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return v_name.startswith(
            f'{FWidgetExtractor.FW_CUSTOM_WIDGET}{FWidgetType.PAGE.value}') \
               or v_name.startswith(FWidgetType.PAGE.value)

    @staticmethod
    def is_node_as_tab_page(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return v_name.startswith(
            f'{FWidgetExtractor.FW_CUSTOM_WIDGET}{FWidgetType.TAB_PAGE.value}') \
               or v_name.startswith(FWidgetType.TAB_PAGE.value)

    @staticmethod
    def is_background(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return v_name.startswith(
            f'{FWidgetExtractor.FW_CUSTOM_WIDGET}{FWidgetType.BACKGROUND.value}') \
               or v_name.startswith(FWidgetType.BACKGROUND.value)

    @staticmethod
    def is_node_as_image(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return f'{FWidgetType.IMAGE.value}_' in v_name \
               or f'{FWidgetType.ICON.value}_' in v_name

    @staticmethod
    def is_node_as_text_view(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return f'{FWidgetType.TEXT.value}_' in v_name \
               or f'{FWidgetType.TEXT_FIELD.value}_' in v_name \
               or v_name == FWidgetType.TEXT.value \
               or v_name == FWidgetType.TEXT_FIELD.value

    @staticmethod
    def is_node_need_dynamic_data(figma_node):
        v_name = figma_node['name'] if 'name' in figma_node else ''
        return FWidgetExtractor.FW_DYNAMIC_DATA in v_name

    @staticmethod
    def find_all_custom_widget_node(current_node, page_nodes):
        if FWidgetExtractor.is_node_as_custom_widget(current_node):
            page_nodes.append(current_node)
        if 'children' in current_node:
            for c_view in current_node['children']:
                FWidgetExtractor.find_all_custom_widget_node(c_view, page_nodes)

    @staticmethod
    def get_resource(current_node: dict,
                     text_dict: OrderedDict[
                                str: [OrderedDict[str:tuple[str, str]]]] = None,
                     text_in_page_dict: OrderedDict[str:tuple[str, str]] = None,
                     hex_color_list: list[str] = None,
                     image_node_dict: dict[str: dict] = None,
                     font_node_list: list[TextStyle] = None):
        """
        :param current_node: Figma node dictionary which we use to look up Flutter resource
        :param text_dict: Keeps all Flutter page strings
        :param text_in_page_dict: Keeps value of list text inside each Flutter page. Will be reset when jump to next page
        :param hex_color_list: Keeps HEX format color of [current_node]
        :param image_node_dict: Keeps images that presented as a [current_node]
        :param font_node_list: Keeps all font styles of [current_node]
        :return:
        """
        if not current_node:
            return

        node_id = current_node['id'] if 'id' in current_node else ''
        node_name = current_node['name'] if 'name' in current_node else ''
        # Find nodes that presented as a Flutter page
        if text_dict is not None:
            if FWidgetExtractor.is_node_as_custom_widget(current_node):
                text_in_page_dict = OrderedDict()
                text_dict[node_name] = text_in_page_dict

        # Find strings and also fonts
        text_related = FWidgetExtractor.text_related(current_node)
        if text_related:
            if text_in_page_dict is not None and not \
                    FWidgetExtractor.is_node_need_dynamic_data(current_node):
                text_in_page_dict[node_id] = node_name, text_related.character
            if font_node_list is not None:
                font_node_list.append(text_related.style)
            if hex_color_list is not None \
                    and text_related.hex_text_color is not None \
                    and text_related.hex_text_color not in hex_color_list:
                hex_color_list.append(text_related.hex_text_color)

        # Find hex colors
        if hex_color_list is not None:
            related_c_hex_list = FWidgetExtractor.color_related(current_node)
            hex_color_list.extend(
                [c for c in related_c_hex_list if
                 c and c not in hex_color_list])

        # Find all nodes contain image
        if image_node_dict is not None:
            if FWidgetExtractor.is_node_as_image(current_node):
                image_node_dict[node_id] = current_node
            # Also lookup background with image
            if FWidgetExtractor.is_background(current_node):
                if len(node_name) > len(FWidgetType.BACKGROUND.value):
                    image_node_dict[node_id] = current_node

        # Recursive child views
        if 'children' in current_node:
            for child_node in current_node['children']:
                FWidgetExtractor.get_resource(
                    current_node=child_node,
                    text_dict=text_dict,
                    text_in_page_dict=text_in_page_dict,
                    hex_color_list=hex_color_list,
                    image_node_dict=image_node_dict,
                    font_node_list=font_node_list
                )


class FTextRelated:
    def __init__(self,
                 character: str,
                 character_style_override: list = None,
                 hex_text_color: str = None,
                 alpha_text_color: int = None,
                 style: TextStyle = None,
                 style_override_table: list = None):
        self.character = character or ''
        self.character_style_override = character_style_override or []
        self.hex_text_color = hex_text_color
        self.alpha_text_color = alpha_text_color or 255
        self.style = style
        self.style_override_table = style_override_table or []

    def present_f_color(self, colors_mapping: dict, property_name: str = None):
        return FWidgetExtractor.present_f_color_from_scratch(
            hex_color=self.hex_text_color,
            alpha=self.alpha_text_color,
            colors_mapping=colors_mapping,
            property_name=property_name
        )


class FRawWidget:
    SCREEN_DESIGN_WIDTH = 375.0
    SCREEN_DESIGN_HEIGHT = 812.0

    def __init__(self,
                 figma_node,
                 v_type: FWidgetType = None,
                 children: list = None):
        self.figma_node = figma_node
        self.type = v_type or FWidgetType.IGNORED
        self.children = children or []
        self.rect = FWidgetExtractor.as_rect(self.figma_node)

    def __eq__(self, other):
        return isinstance(other, __class__) \
               and other.id_view == self.id_view

    def __hash__(self):
        return id(self.id_view)

    @property
    def id_view(self):
        return self.figma_node['id'] if 'id' in self.figma_node else ''

    @property
    def raw_name(self):
        return self.figma_node['name'] if 'name' in self.figma_node else ''

    @property
    def full_name(self):
        if self.is_custom_widget:
            return self.raw_name.replace(FWidgetExtractor.FW_CUSTOM_WIDGET, '')
        return self.raw_name

    @property
    def is_custom_widget(self):
        return self.raw_name.startswith(FWidgetExtractor.FW_CUSTOM_WIDGET)

    @property
    def present(self):
        v_type = f'[{self.type.name}]'
        v_id = f'[{self.id_view}]'
        return f'{v_type}{v_id} {self.full_name}'

    def optimize_view_type(self):
        if self.type == FWidgetType.STACK:
            # Firstly, remove mask node if this stack has, also collect size/radius from mask
            mask_node = None
            for child in self.children:
                if child.figma_node.get('isMask', False) is True:
                    mask_node = child
                    break
            if mask_node:
                self.rect = FWidgetExtractor.as_rect(mask_node.figma_node)
                self.children.remove(mask_node)
            # Check if this STACK has first child showing as its BACKGROUND
            if not self.children:
                return
            first_child = self.children[0]
            if first_child.type != FWidgetType.BACKGROUND \
                    or not first_child.rect.is_background_of(self.rect):
                return
            remain_children = self.children[1:]
            if len(remain_children) < 2:
                self.children = remain_children
            else:
                # Create new stack to contains remaining children
                # and make current stack contains that ones as an only child
                new_stack = FRawWidget(
                    figma_node=self.figma_node,
                    v_type=FWidgetType.STACK,
                    children=remain_children
                )
                self.children = [new_stack]
            # Then we convert this stack to container immediately
            self.figma_node = first_child.figma_node
            self.type = FWidgetType.CONTAINER
            self.rect = FWidgetExtractor.as_rect(self.figma_node)



import hashlib


class Encoder:

    def encode(self, alg, string):
        if alg == 'md5':
            return self.encode_md5(string)
        return string

    @staticmethod
    def encode_md5(string):
        return hashlib.md5(string.encode()).hexdigest()



# coding=utf-8
import re
import time


def auto_str(cls):
    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    cls.__str__ = __str__
    return cls


def create_file_name(class_name) -> str:
    if is_camel_case(class_name):
        upper_case_letters = re.findall('[A-Z][^A-Z]*', class_name)
        return '_'.join(upper_case_letters).lower()
    else:
        print(
            'Invalid filename. Filename need to be camel string.\nFor '
            'example: MyCamelCase')
        exit(1)


def is_camel_case(st) -> str:
    return st != st.lower() and st != st.upper() and "_" not in st


def lower_first_letter(st) -> str:
    return st[0].lower() + st[1:]


def upper_first_letter(st) -> str:
    return st[0].upper() + st[1:]


def snake_to_camel(st) -> str:
    components = st.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def snake_to_pascal(st) -> str:
    components = st.split('_')
    return ''.join(x.title() for x in components)


def to_words(st) -> str:
    return ' '.join(to_snake_case(st).split('_')).lstrip()


def to_snake_case(st) -> str:
    result = []
    for i in range(len(st)):
        char = st[i]
        if char == ' ' or char == '_':
            if (i < len(st) - 1) and (st[i + 1] == '_'
                                      or st[i + 1] == ' '
                                      or st[i + 1].isupper()):
                continue
            result.append('_')
        elif char.isupper():
            result.append(f'{"_" if i > 0 else ""}{char.lower()}')
        else:
            result.append(char)
    return ''.join([c for c in result])


def to_camel_case(st) -> str:
    return snake_to_camel(to_snake_case(st))


def to_pascal_case(st) -> str:
    return snake_to_pascal(to_snake_case(st))


def plural_to_singular(st) -> str:
    if st.endswith('ies'):
        if len(st) > 3:
            if st[-4] not in 'aeiou':
                return st[0:-3] + 'y'
    elif st.endswith('es'):
        if st[-3] in 'sxo':
            return st[0:-2]
        elif st[-4:-2] == 'ch' or st[-4:-2] == 'sh':
            return st[0:-2]
        else:
            return st[0:-1]
    elif st.endswith('s'):
        if len(st) > 3:
            return st[:-1]
    return st


def is_number(st) -> bool:
    try:
        float(st)
        return True
    except ValueError:
        pass

    try:
        import unicodedata
        unicodedata.numeric(st)
        return True
    except (TypeError, ValueError):
        pass

    return False


def is_time_format(st):
    try:
        time.strptime(st, '%H:%M')
        return True
    except ValueError:
        return False



from kntgen.model.flutter.flutter_base_widgets import *


class FConverter:
    @staticmethod
    def convert(raw_widget: FRawWidget):
        """
        :return: Flutter Specific Widget
        """
        f_widget = None
        if raw_widget.type == FWidgetType.NEW_WIDGET:
            f_widget = FCustom(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.PAGE:
            f_widget = FPage(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.TAB_PAGE:
            f_widget = FTabPage(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.CONTAINER \
                or raw_widget.type == FWidgetType.BACKGROUND:
            f_widget = FContainer(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.COLUMN:
            f_widget = FColumn(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.ROW:
            f_widget = FRow(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.WRAP:
            f_widget = FWrap(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.WRAP_VERTICAL:
            f_widget = FWrap(raw_widget.figma_node, is_horizontal=False)
        if raw_widget.type == FWidgetType.STACK:
            f_widget = FStack(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.LIST_VIEW:
            f_widget = FList(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.LIST_VIEW_HORIZONTAL:
            f_widget = FList(raw_widget.figma_node, is_horizontal=True)
        if raw_widget.type == FWidgetType.GRID_VIEW:
            f_widget = FGrid(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.GRID_VIEW_HORIZONTAL:
            f_widget = FGrid(raw_widget.figma_node, is_horizontal=True)
        if raw_widget.type == FWidgetType.APP_BAR:
            f_widget = FAppbar(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.TOOL_BAR:
            f_widget = FToolbar(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.BOTTOM_BAR:
            f_widget = FBottomBar(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.TEXT:
            f_widget = FText(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.TEXT_FIELD:
            f_widget = FTextField(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.IMAGE \
                or raw_widget.type == FWidgetType.ICON:
            f_widget = FImage(raw_widget.figma_node)
        if raw_widget.type == FWidgetType.DIVIDER:
            f_widget = FDivider(raw_widget.figma_node)
        if not f_widget:
            f_widget = FWidget(raw_widget.figma_node)
        # Recursive to convert this raw widget's children
        for child_raw_w in raw_widget.children:
            f_child_widget = FConverter.convert(child_raw_w)
            if f_child_widget:
                f_child_widget.parent = f_widget
                # Row/Column will has relative size of it's wrapper
                if child_raw_w.type == FWidgetType.ROW:
                    f_child_widget.rect.left = f_widget.rect.left
                    f_child_widget.rect.right = f_widget.rect.right
                if child_raw_w.type == FWidgetType.COLUMN:
                    f_child_widget.rect.top = f_widget.rect.top
                    f_child_widget.rect.bottom = f_widget.rect.bottom
                f_widget.children.append(f_child_widget)
        return f_widget

    @staticmethod
    def generate_widgets(current_widget,
                         gen_content_dict: dict,
                         strings_mapping: dict,
                         colors_mapping: dict,
                         images_mapping: dict):
        if current_widget is None or gen_content_dict is None:
            return
        if not isinstance(current_widget, FCustom):
            return
        content = current_widget.generate(
            strings_mapping=strings_mapping,
            colors_mapping=colors_mapping,
            images_mapping=images_mapping
        )
        gen_content_dict[
            f'{current_widget.generate_relative_file_path()}'] = content


class FPage(FCustom):
    def generate_build(self,
                       strings_mapping: dict,
                       colors_mapping: dict,
                       images_mapping: dict):
        # FIXME: check condition
        bottom_bar_present = f"""
        bottomNavigationBar: BottomNav(
                starterTabIndex: _currentIndexTab,
                onChoseTab: _onChoseTab,
              ), """ \
            if self.has_bottom_navigation_bar() else ''
        tab_present = f"""CustomBB()""" if self.has_bottom_navigation_bar() else \
            super().generate_build(
                strings_mapping=strings_mapping,
                colors_mapping=colors_mapping,
                images_mapping=images_mapping
            )
        return f"""
BlackStatusBarPage(
    page: Scaffold({bottom_bar_present}
        body: {tab_present},
    ),
)"""

    def generate(self,
                 strings_mapping: dict,
                 colors_mapping: dict,
                 images_mapping: dict):
        custom_generate = super().generate(strings_mapping=strings_mapping,
                                           colors_mapping=colors_mapping,
                                           images_mapping=images_mapping)
        return f"""
import '../custom/custom_status_bar_page.dart';
{custom_generate}
"""

    def has_bottom_navigation_bar(self):
        for child in self.children or []:
            if FWidgetType.BOTTOM_BAR.value in child.w_name:
                return True
        return False

    def generated_file_name(self):
        return f'{self.w_name}_page'

    def generate_container_folder(self):
        return 'pages'

    @property
    def w_class(self):
        return to_pascal_case(self.w_name)


class FTabPage(FCustom):
    def generated_file_name(self):
        page_name = self._interpolate_page_name
        tab_name = self.w_name.replace(f'_{page_name}', '') \
            if f'_{page_name}' in self.w_name else self.w_name
        return f'tab_{tab_name}'

    def generate_container_folder(self):
        return f'pages/{self._interpolate_page_name}/tabs'

    @property
    def w_class(self):
        return f'Tab{to_pascal_case(self.w_name)}'

    @property
    def _interpolate_page_name(self):
        return to_snake_case(self.w_raw_name).split('_')[-1]


class FAppbar(FWidget):
    @property
    def w_class(self):
        return 'KntDefaultAppBar'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        return [
            'titlePage: \'{empty_title}\''
        ]


class FToolbar(FWidget):
    @property
    def w_class(self):
        return 'KntDefaultToolBar'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        return [
            'titlePage: \'{empty_title}\''
        ]


class FBottomBar(FWidget):
    @property
    def w_class(self):
        return 'KntBottomNavScaffold'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        return [
            'onTabChanged: ValueNotifier<int>(0)',
            'appBar: KntAppBar(titlePage: \'{empty_title}\')',
            'icons: []',
            'tabs: []'
        ]


class FContainer(FWidget):
    @property
    def w_class(self):
        return 'Container'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        width, height = self.w_size_present
        # Ignore size if parent widget has type FCustom
        if isinstance(self.parent, FCustom):
            width = ''
            height = ''
        # Calculate insets if needed
        margin = ''
        if isinstance(self.parent, FViewGroup):
            # Merge Flutter Padding#padding with this Container#margin
            margin = FWidget.extract_edge_insets(
                self.margin,
                specific_edges=self.parent.needed_child_edge_insets,
                inset_type=InsetType.MARGIN
            )
        padding = ''
        if self.children:
            padding = FWidget.extract_edge_insets(
                self.padding,
                specific_edges=[Edge.LEFT, Edge.TOP, Edge.RIGHT, Edge.BOTTOM]
            )
        return [
            width,
            height,
            margin,
            padding,
            self._decoration(
                colors_mapping=colors_mapping,
                images_mapping=images_mapping
            ),
            self.generate_child(
                strings_mapping=strings_mapping,
                colors_mapping=colors_mapping,
                images_mapping=images_mapping
            )
        ]

    def generate_wrapper(self, generated_content):
        return generated_content

    def _decoration(self, colors_mapping: dict, images_mapping: dict):
        deco_plist = [
            self._deco_background_color(colors_mapping),
            self._deco_radius(),
            self._deco_stroke(colors_mapping),
            self._deco_shadow(colors_mapping),
            self._deco_image(images_mapping)
        ]
        deco_plist_present = ''.join(f'{d},\n' for d in deco_plist if d)
        return f'decoration: BoxDecoration({deco_plist_present})' \
            if deco_plist_present else ''

    def _deco_image(self, images_mapping: dict):
        image_path = FWidget.extract_image_path(self.w_id, images_mapping)
        return f'image: DecorationImage(image: AssetImage({image_path}), ' \
               f'fit: BoxFit.cover,)' if image_path else None

    def _deco_background_color(self, colors_mapping: dict):
        return FWidget.extract_colors(self.figma_node, colors_mapping)

    def _deco_stroke(self, colors_mapping: dict):
        return FWidget.extract_border(self.figma_node, colors_mapping)

    def _deco_radius(self):
        return FWidget.extract_border_radius(self.figma_node)

    def _deco_shadow(self, colors_mapping: dict):
        return FWidget.extract_shadows(self.figma_node, colors_mapping)


class FText(FWidget):
    SPECIAL_CHARS = '\'$'
    EMPTY_TEXT = '{empty_text}'

    def __init__(self, figma_node, children: list = None):
        super().__init__(figma_node, children)
        self.text_related = FWidgetExtractor.text_related(figma_node)

    @property
    def w_class(self):
        # TODO: Rich text if self.text_related has not None character_style_override
        return 'Text'

    @property
    def is_flex_size(self):
        return True

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        if not self.text_related:
            return [FText.EMPTY_TEXT]
        style = self.text_related.style
        character = ''.join(
            [c for c in self.text_related.character if
             c not in FText.SPECIAL_CHARS])
        character = f'{character!r}'
        if self.w_id in strings_mapping:
            localized_text = strings_mapping[self.w_id]
            if localized_text:
                character = localized_text
        color_present = self.text_related.present_f_color(colors_mapping)
        if color_present:
            color_present = f'{color_present},'
        italic_present = '\nfontStyle: FontStyle.italic,' if style.italic else ''
        return [
            character,
            f"""style: TextStyle( {color_present}
    fontFamily: FontFamily.{to_camel_case(style.font_family)},
    fontSize: {style.font_size}.sp,
    fontWeight: FontWeight.w{style.font_weight}, {italic_present}
    height: {style.line_height_px}.sp / {style.font_size}.sp,
)"""
        ]


class FTextField(FText):
    @property
    def w_class(self):
        return 'KntSimpleEditText'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        if not self.text_related:
            return []
        style = self.text_related.style
        character = ''.join(
            [c for c in self.text_related.character if
             c not in FText.SPECIAL_CHARS])
        character = f'{character!r}'
        if self.w_id in strings_mapping:
            localized_text = strings_mapping[self.w_id]
            if localized_text:
                character = localized_text
        color_present = self.text_related.present_f_color(colors_mapping,
                                                          property_name='hintTextColor')
        return [
            f'hintText: {character}',
            color_present,
            f'fontSize: {style.font_size}.sp'
        ]


class FImage(FWidget):
    @property
    def w_class(self):
        return 'KntImage.assets'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        width, height = self.w_size_present
        return [
            self._image_url(images_mapping),
            width,
            height,
            self._border_radius()
        ]

    def _image_url(self, images_mapping: dict):
        image_path = FWidget.extract_image_path(self.w_id, images_mapping)
        return f'uri: {image_path}' if image_path else None

    def _border_radius(self):
        return FWidget.extract_border_radius(self.figma_node)


class FDivider(FWidget):
    @property
    def w_class(self):
        return 'const KntDivider'


class FColumn(FViewGroup):
    @property
    def w_class(self):
        return 'Column'

    @property
    def main_alignment(self):
        return self.y_alignment

    @property
    def cross_alignment(self):
        return self.x_alignment

    @property
    def direction(self):
        return Direction.VERTICAL

    def generate_direction(self) -> str:
        return None


class FRow(FViewGroup):
    @property
    def w_class(self):
        return 'Row'

    @property
    def main_alignment(self):
        return self.x_alignment

    @property
    def cross_alignment(self):
        return self.y_alignment

    @property
    def direction(self):
        return Direction.HORIZONTAL

    def generate_direction(self) -> str:
        return None


class FWrap(FViewGroup):
    def __init__(self, figma_node, children: list = None, is_horizontal=True):
        super().__init__(figma_node, children)
        self.is_horizontal = is_horizontal

    @property
    def w_class(self):
        return 'Wrap'

    def generate_direction(self) -> str:
        return 'direction: Axis.vertical' if self.direction == Direction.VERTICAL else None

    def generate_alignments(self) -> list[str]:
        return []


class FStack(FWidget):
    @property
    def w_class(self):
        return 'Stack'

    def generate_child_wrapper(self, child_widget, generated_child_content):
        if not generated_child_content:
            return generated_child_content
        positioned = FStack.calculate_child_position(child_widget)
        return f'Positioned({positioned},child:{generated_child_content},)' \
            if positioned else generated_child_content

    @staticmethod
    def calculate_child_position(child_widget):
        if not child_widget or not child_widget.margin \
                or len(child_widget.margin) < 4:
            return ''
        # Fetch margin from this child
        margin = [v for v in child_widget.margin.values()]
        # Present margin
        horizontal_position = f'left: {margin[0]}.w' if margin[0] < margin[
            2] else f'right: {margin[2]}.w'
        if child_widget.rect.width > child_widget.parent.rect.width \
                * (1 - FWidget.POSITIONED_STACK_OFFSET_RATIO):
            horizontal_position = f'left: {margin[0]}.w, right: {margin[2]}.w'
        vertical_position = f'top: {margin[1]}.w' if margin[1] < margin[3] \
            else f'bottom: {margin[3]}.w'
        return ','.join([
            horizontal_position,
            vertical_position
        ])


class FList(FViewGroupFixedStyle):
    @property
    def w_class(self):
        return 'ListView.builder'


class FGrid(FViewGroupFixedStyle):
    @property
    def w_class(self):
        return 'GridView.builder'

    def generate_alignments(self) -> list[str]:
        ratio = ''
        child_size = self.presented_child_size
        if child_size:
            ratio = f'childAspectRatio: {child_size[0]} / {child_size[1]},'
        return [
            'gridDelegate: SliverGridDelegateWithFixedCrossAxisCount('
            'crossAxisCount: 2,'
            'mainAxisSpacing: 8.w,'
            'crossAxisSpacing: 8.w,'
            f'{ratio}'
            ')'
        ]



import subprocess

from kntgen.printer import *

FLUTTER_FORMAT = 'flutter format .'
FLUTTER_GEN = 'flutter pub run build_runner build --delete-conflicting-outputs'
FLUTTER_GEN_LOCALE = 'flutter pub run easy_localization:generate ' \
                     '-S assets/translations -f keys -o locale_keys.g.dart'


def bash_execute(command: str, need_print=True):
    if need_print:
        info_printer(f'Running command line [{command}] ...')
    process = subprocess.Popen((command or FLUTTER_FORMAT).split(),
                               stdout=subprocess.PIPE)
    output, error = process.communicate()
    if error:
        print(f'Occur an error while running command line: [{error}]')



# coding=utf-8


class Command:
    TAB_LENGTH = '<15'

    @classmethod
    def description(cls):
        return ''

    @classmethod
    def name(cls):
        return ''

    @classmethod
    def class_name(cls):
        return cls.__name__

    @classmethod
    def long_description(cls):
        return f'{cls.name():{Command.TAB_LENGTH}}{cls.description()}'


class CommandOption:

    def __init__(self, options):
        self._options = options

    def __getattr__(self, name):
        cls = type(self)
        if name in self._options:
            return self._options[name]
        else:
            raise AttributeError(
                f'{cls.__name__!r} object has no attribute {name!r}')



from kntgen.bloc.bloc_template import BlocTemplate
from kntgen.cmd import Command, CommandOption
from kntgen.constants import SUPPORTED_SOURCE_FORMAT
from kntgen.file_helpers import get_flutter_project_package
from kntgen.model.bloc import Bloc
from kntgen.model.exception import DefinedErrorMessage


class BlocCommand(Command):
    def __init__(self, options):
        super(BlocCommand, self).__init__()
        # For future usage
        self.options = CommandOption(options)
        self.project_package = get_flutter_project_package()

    def create_files(self, content_file: str, file_extension: str):
        try:
            if file_extension == SUPPORTED_SOURCE_FORMAT[0]:
                self._create_files_from_json(content_file)
            elif file_extension == SUPPORTED_SOURCE_FORMAT[1]:
                self._create_files_from_yaml(content_file)
        except Exception as e:
            DefinedErrorMessage.print_error(e)
            exit(1)

    def _create_files_from_json(self, json_str: str):
        bloc = BlocTemplate(
            self.project_package,
            Bloc.from_json(json_str)
        )
        bloc.create_files()

    def _create_files_from_yaml(self, yaml_stream):
        bloc = BlocTemplate(
            self.project_package,
            Bloc.from_yaml(yaml_stream)
        )
        bloc.create_files()



import json
from collections import OrderedDict
from typing import List

from jinja2 import Environment, PackageLoader

from kntgen.collection_helpers import merged_non_empty
from kntgen.constants import *
from kntgen.str_helpers import *


class JSON(object):
    IMPORT_HEADER = 'import \'package:json_annotation/json_annotation.dart\';\n'

    class Property(object):
        def __init__(self, raw_name: str, name: str, type_name: str):
            self.raw_name = raw_name
            self.name = name
            self.type_name = type_name

        def __eq__(self, other):
            return isinstance(other, __class__) \
                   and other.raw_name == self.raw_name \
                   and other.name == self.name \
                   and other.type_name == self.type_name

        def __hash__(self):
            return id(f'{self.raw_name}_{self.name}_{self.type_name}')

        @property
        def is_user_type(self):
            return self.is_optional \
                   and self.original_type_name not in DART_TYPES_DEFAULT_VALUES

        @property
        def is_optional(self):
            return self.type_name.endswith('?')

        @property
        def is_list(self):
            return self.type_name.startswith('List')

        @property
        def is_array(self):
            return self.type_name.startswith('[')

        @property
        def is_date(self):
            return self.original_type_name == 'DateTime'

        @property
        def is_dart_type(self):
            return self.type_name in DART_TYPES_DEFAULT_VALUES

        @property
        def original_type_name(self):
            origin_name = self.type_name
            if self.is_list:
                origin_name = origin_name.removeprefix('List')
            return ''.join(c for c in origin_name if c not in '?[]<>')

        @property
        def value(self):
            if self.is_optional:
                value = 'null'
            elif self.is_array or self.is_list:
                value = 'const []'
            elif self.type_name in DART_TYPES_DEFAULT_VALUES:
                value = DART_TYPES_DEFAULT_VALUES[self.type_name]
            else:
                value = 'null'
            if self.type_name.startswith('String'):
                value = f'{value!r}'
            return value

    class Model(object):
        def __init__(self, name: str,
                     properties: set = None):
            self.name = name
            self.properties = properties or {}
            self.env = Environment(
                loader=PackageLoader('kntgen_templates', 'json_to_dart'),
                trim_blocks=True,
                lstrip_blocks=True
            )

        def __eq__(self, other):
            return isinstance(other, __class__) and other.name == self.name

        def __hash__(self):
            return id(f'{self.name}_{len(self.properties)}')

        def create_file_from_template(self):
            template = self.env.get_template('json.dart')
            content = template.render(
                model_name=self.name,
                properties=self.properties
            )
            return content

    def __init__(self, model_name: str, json_str: str):
        self.model_name = model_name
        self.json_str = json_str

    def create_model_files(self):
        models = list(
            reversed(JSON._remove_duplicated_models(self._create_models())))
        model_name_snake_case = to_snake_case(self.model_name)
        import_part = f'\npart \'{model_name_snake_case}.g.dart\';\n'
        output = '\n'.join(
            [model.create_file_from_template() for model in models]
        )
        output = JSON.IMPORT_HEADER + import_part + output + '\n'
        return output

    def _create_models(self):
        try:
            dictionary = json.loads(self.json_str,
                                    object_pairs_hook=OrderedDict)
            models = []
            self._extract_to_dart_model(self.model_name, dictionary, models)
            return models
        except Exception as e:
            print(f'The JSON is invalid! {e.__repr__()}')
            exit(1)

    def _extract_to_dart_model(self,
                               model_name: str,
                               dictionary: dict,
                               models: List[Model]):
        properties = []
        if type(dictionary).__name__ == 'list':
            if len(dictionary) > 0:
                if type(dictionary[0]).__name__ == 'OrderedDict':
                    most_key_dict = dictionary[0]
                    # Join all JSON object to find all properties
                    for sub_dict in dictionary[1:]:
                        most_key_dict = merged_non_empty(
                            most_key_dict,
                            sub_dict
                        )
                    self._extract_to_dart_model(
                        model_name,
                        most_key_dict,
                        models
                    )
            return
        for key, value in dictionary.items():
            var_name = snake_to_camel(key)
            type_name = type(value).__name__
            if type_name == 'OrderedDict':
                var_type = snake_to_pascal(key)
                self._extract_to_dart_model(var_type, value, models)
            elif type_name == 'list':
                singular_var_name = plural_to_singular(var_name)
                if len(value) > 0:
                    sub_var_type = type(value[0]).__name__
                    if sub_var_type == 'OrderedDict':
                        most_key_dict = value[0]
                        # Join all JSON object to find all properties
                        for sub_dict in value[1:]:
                            most_key_dict = merged_non_empty(
                                most_key_dict,
                                sub_dict
                            )
                        var_type = f'List<{singular_var_name.title()}?>'
                        self._extract_to_dart_model(
                            singular_var_name.title(),
                            most_key_dict,
                            models
                        )
                    else:
                        var_type = f'List<{PYTHON_TO_DART_TYPES[sub_var_type]}?>'
                else:
                    var_type = 'List'
            else:
                var_type = PYTHON_TO_DART_TYPES[type_name]
                if var_type == 'String':
                    if re.match(DATE_REGEX, value):
                        var_type = 'DateTime?'
                if var_type == 'dynamic':
                    lower_var_name = var_name.lower()
                    # Filter return type with preset common name
                    if lower_var_name.endswith('url') \
                            or lower_var_name.endswith('key') \
                            or lower_var_name.endswith('token') \
                            or 'name' in lower_var_name \
                            or 'email' in lower_var_name:
                        var_type = 'String'
                    elif lower_var_name.endswith('at') \
                            or 'time' in lower_var_name \
                            or 'date' in lower_var_name \
                            or 'birthday' in lower_var_name:
                        var_type = 'DateTime?'
                    elif lower_var_name.endswith('id'):
                        var_type = 'num'
            if var_type not in DART_TYPES_DEFAULT_VALUES \
                    and not var_type.startswith('List'):
                var_type = var_type + '?'
            json_property = JSON.Property(key, var_name, var_type)
            properties.append(json_property)
        model = JSON.Model(model_name, properties)
        models.append(model)

    @staticmethod
    def _remove_duplicated_models(models: list[Model]):
        not_dup_models = []
        total_models = len(models)
        for i in range(total_models):
            model = models[i]
            if model in not_dup_models:
                continue

            for j in range(i + 1, total_models):
                if model == models[j]:
                    # Supplement lacked properties from duplicated models
                    # into existed model
                    model.properties = list(
                        set(model.properties) | set(models[j].properties))

            not_dup_models.append(models[i])
        return not_dup_models



from collections import Mapping, OrderedDict


def merged_non_empty(d1: dict, d2: dict) -> dict:
    return {
        k: (d1.get(k) or d2.get(k))
        for k in set(d1) | set(d2)
    }


def insert_at_pos_ordered_dict(dic,
                               current_key_dict=None,
                               item=None,
                               key_to_insert_after=None):
    if item is None:
        return
    d = OrderedDict()
    insert_at_first = key_to_insert_after is None
    if insert_at_first:
        d[current_key_dict] = item
    for dic_k, dic_v in dic.items():
        d[dic_k] = dic_v
        if not insert_at_first and dic_k == key_to_insert_after:
            d[current_key_dict] = item
    # Replace with new ordered
    dic.clear()
    dic.update(d)


def ordered_dict_merge(result_d: OrderedDict,
                       complement_d: OrderedDict):
    """ Recursive OrderedDict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, ordered_dict_merge recurse down into dicts nested
    to an arbitrary depth, updating keys. The ``complement_d`` is merged into
    ``result_d``.

    :param result_d: dict onto which the merge is executed
    :param complement_d: dct merged into dct
    :return: None
    """
    for k, v in complement_d.items():
        if k in result_d and isinstance(result_d[k], OrderedDict) \
                and isinstance(v, OrderedDict):
            ordered_dict_merge(result_d[k], v)
        else:
            result_d[k] = v


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurse down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, v in merge_dct.iteritems():
        if k in dct and isinstance(dct[k], dict) \
                and isinstance(merge_dct[k], Mapping):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]



# coding=utf-8
from kntgen.file_helpers import make_content_output_file
from kntgen.model.bloc import Bloc
from kntgen.printer import beauty_printer
from kntgen.str_helpers import *
from kntgen.template import Template


@auto_str
class BlocTemplate(Template):
    def __init__(self, package_name: str, bloc: Bloc):
        super().__init__('bloc')
        self.name_lower = create_file_name(bloc.name)
        self.package_name = package_name
        self.bloc = bloc
        self.content_print_out = []

    def extract_content_from_template(self, template):
        return template.render(
            bloc=self.bloc,
            package_name=self.package_name
        )

    def create_files(self):
        self._create_bloc()
        self._create_bloc_test()
        beauty_printer(title='Successfully created BLoC related files',
                       content=self.content_print_out)

    def _create_bloc(self):
        scene_name = 'bloc'
        file_path, additional_info = self.create_file_from_template(
            class_name=f'{self.name_lower}_{scene_name}',
            template_file=f'{scene_name}.dart',
            output_path='lib/bloc'
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))

    def _create_bloc_test(self):
        scene_name = 'bloc_test'
        file_path, additional_info = self.create_file_from_template(
            class_name=f'{self.name_lower}_{scene_name}',
            template_file=f'{scene_name}.dart',
            output_path='test/bloc'
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))



import yaml

from kntgen.str_helpers import auto_str

DEFAULT_PATH_ASSETS_IMAGES = './assets/images'
DEFAULT_PATH_ASSETS_COLORS = './assets/colors'
DEFAULT_PATH_ASSETS_FONTS = './assets/fonts'
DEFAULT_PATH_ASSETS_TRANSLATIONS = './assets/translations'
DEFAULT_PLACEHOLDER_TRANSLATE = 'need_to_translate'
DEFAULT_LOCALE = 'en'
SVG_MAXIMUM_SIZE = 96
DEFAULT_SCREEN_WIDTH = 375
DEFAULT_SCREEN_HEIGHT = 812


@auto_str
class FigmaInfo:
    def __init__(self, **kwargs):
        self.access_token = kwargs.get('access_token', '')
        self.file_id = kwargs.get('file_id', '')
        self.node_id = kwargs.get('node_id', '')
        self.maximum_svg_size_in_px = kwargs.get('maximum_svg_size_in_px', 128)
        self.screen_width = kwargs.get('screen_width', DEFAULT_SCREEN_WIDTH)
        self.screen_height = kwargs.get('screen_height', DEFAULT_SCREEN_HEIGHT)


@auto_str
class GoogleFontInfo:
    def __init__(self, **kwargs):
        self.api_key = kwargs.get('api_key', '')
        self.basic_font = kwargs.get('basic_font', False)


@auto_str
class GoogleTranslateInfo:
    def __init__(self, **kwargs):
        self.main_locale = kwargs.get('main_locale', DEFAULT_LOCALE)
        self.placeholder = kwargs.get('placeholder',
                                      DEFAULT_PLACEHOLDER_TRANSLATE)


@auto_str
class GoogleInfo:
    def __init__(self, **kwargs):
        self.font = GoogleFontInfo(
            **kwargs['font']) if 'font' in kwargs else GoogleFontInfo()
        self.translate = GoogleTranslateInfo(**kwargs[
            'translate']) if 'translate' in kwargs else GoogleTranslateInfo()


@auto_str
class AssetsInfo:
    def __init__(self, **kwargs):
        self.images = kwargs.get('images', DEFAULT_PATH_ASSETS_IMAGES)
        self.colors = kwargs.get('colors', DEFAULT_PATH_ASSETS_COLORS)
        self.fonts = kwargs.get('fonts', DEFAULT_PATH_ASSETS_FONTS)
        self.translations = kwargs.get('translations',
                                       DEFAULT_PATH_ASSETS_TRANSLATIONS)


@auto_str
class UiConfig:
    def __init__(self, **kwargs):
        self.figma = FigmaInfo(
            **kwargs['figma']) if 'figma' in kwargs else FigmaInfo()
        self.google = GoogleInfo(
            **kwargs['google']) if 'google' in kwargs else GoogleInfo()
        self.assets = AssetsInfo(
            **kwargs['assets']) if 'assets' in kwargs else AssetsInfo()

    @classmethod
    def from_yaml(cls, yaml_stream):
        yaml_model = yaml.safe_load(yaml_stream)
        return cls(**yaml_model)



from kntgen.constants import *
from kntgen.file_helpers import *
from kntgen.model.exception import *
from kntgen.pasteboard import *


def handle_base_parser(parser, args, supported_sources=SUPPORTED_SOURCES):
    """
    :param parser: parsers from [subcmd] package
    :param args: arguments from [subcmd] package, must include argument [--source] and [--path]
    :param supported_sources: Supported resources to handle
    :return: A tuple of (current args, content of the file which extracted from args, format of content)
    """
    extracted_args = parser.parse_args(args)
    source_type = None
    if extracted_args.source:
        source_type = extracted_args.source[0]
        if source_type not in supported_sources:
            parser.error(DefinedErrorMessage.ERROR_SUPPORT_SOURCE_INPUT)

    content_file = ''
    file_extension = ''
    if not source_type or source_type == SUPPORTED_SOURCES[0]:
        # Use [pasteboard] when [-s] is not defined
        content_file = pasteboard_read()
        if validate_json_content(content_file):
            file_extension = SUPPORTED_SOURCE_FORMAT[0]
        elif validate_yaml_content(content_file):
            file_extension = SUPPORTED_SOURCE_FORMAT[1]
    else:
        # Required file path when [-s] is [file]
        if not extracted_args.path:
            parser.error(DefinedCmd.Required.INPUT_FILE)
        file_path = extracted_args.path[0]
        if not os.path.exists(file_path):
            parser.error(DefinedErrorMessage.FILE_NOT_FOUND)
        file_extension = os.path.splitext(file_path)[1].removeprefix('.')
        if file_extension not in SUPPORTED_SOURCE_FORMAT:
            parser.error(DefinedErrorMessage.ERROR_SUPPORT_FORMAT)
        try:
            with open(file_path, 'r') as file:
                content_file = file.read()
        except (IOError, OSError) as e:
            DefinedErrorMessage.print_error(e)
            exit(1)

    if not file_extension or file_extension.isspace():
        parser.error(DefinedErrorMessage.ERROR_SUPPORT_FORMAT)
    return extracted_args, content_file, file_extension


def create_base_argument(parser,
                         description: str,
                         target: str,
                         supported_sources: list = None):
    parser.description = description
    parser.add_argument(
        '-s', '--source',
        nargs=1,
        choices=supported_sources or SUPPORTED_SOURCES,
        required=False,
        help=DefinedCmd.Argument.help_source(target)
    )
    parser.add_argument(
        '-p', '--path',
        nargs=1,
        required=False,
        help=DefinedCmd.Argument.help_path(target)
    )


def is_set(arg_name):
    if arg_name in sys.argv:
        return True
    return False



from kntgen.bash_helpers import *
from kntgen.model.flutter.flutter_raw_widget import *
from kntgen.model.flutter.flutter_specific_widgets import FConverter
from kntgen.ui.flutter_generator import *
from kntgen.ui.ui_external_source import *


class FParser:
    def __init__(self, parent_node, figma_py, ui_config: UiConfig):
        self.parent_node = parent_node
        self.figma_py = figma_py
        self.ui_config = ui_config
        self.figma_config = ui_config.figma
        self.google_config = ui_config.google
        self.assets_config = ui_config.assets
        self.file_id = self.figma_config.file_id


# -------------------------------------------------------------------------
# FLUTTER STRING
# -------------------------------------------------------------------------
class FStringParser(FParser):
    def generate(self,
                 folder: str = None,
                 pivot_locale: str = None,
                 need_print=True):
        # Fetch strings
        text_dict = OrderedDict()
        FWidgetExtractor.get_resource(self.parent_node,
                                      text_dict=text_dict,
                                      text_in_page_dict=OrderedDict())
        # Generate
        FStringParser.generate_flutter_source(strings=text_dict,
                                              locale_folder_path=folder,
                                              pivot_locale=pivot_locale,
                                              need_print=need_print)

    @staticmethod
    def generate_flutter_source(strings: OrderedDict = None,
                                locale_folder_path=None,
                                pivot_locale=None,
                                need_print=True):
        """
        Generate localization files. Get existed content from pivot file then translate into the other files. Just support JSON format

        :param strings: Dict contains locale keys
        :param locale_folder_path: Path to folder which contains locale files
        :param pivot_locale: File for defined which language is default
        :param need_print: Whether need printout generated files
        :return:
        """
        if strings is None:
            error_printer('Could not found any strings!')
            return
        warning_printer('Generating strings...')
        if need_print:
            for page, ordered_d in strings.items():
                if not ordered_d:
                    error_printer(f'Screen {page!r} has no text!',
                                  emphasis=False)
                    continue
                beauty_printer(
                    title=f'Text in screen {page!r}',
                    content=[f'[{text_id}] {text[1]}' for text_id, text in
                             ordered_d.items()],
                    max_char_inline=60
                )
        FGenerator.generate_json_locale_file(strings,
                                             locale_folder_path=locale_folder_path,
                                             pivot_locale=pivot_locale,
                                             need_print=need_print)


# -------------------------------------------------------------------------
# FLUTTER COLOR
# -------------------------------------------------------------------------
class FColorParser(FParser):
    def generate(self, folder=None, need_print=True):
        # Fetch colors
        hex_color_list = []
        FWidgetExtractor.get_resource(self.parent_node,
                                      hex_color_list=hex_color_list)
        # Fetch named colors
        named_colors = fetch_named_colors(hex_color_list or [])
        # Generate
        FColorParser.generate_flutter_source(named_colors=named_colors,
                                             folder=folder,
                                             need_print=need_print)

    @staticmethod
    def generate_flutter_source(named_colors: dict[str: NamedColor] = None,
                                folder: str = None,
                                need_print=True):
        if not named_colors:
            error_printer('Could not found any colors!')
            return
        warning_printer('Generating colors...')
        if need_print:
            beauty_print_named_colors(named_colors)
        FGenerator.generate_color_file(named_colors=named_colors,
                                       folder=folder,
                                       need_print=need_print)


# -------------------------------------------------------------------------
# FLUTTER IMAGE
# -------------------------------------------------------------------------
class FImageParser(FParser):
    def generate(self,
                 folder=None,
                 svg_maximum_size=None,
                 need_print=True):
        # Fetch images information
        image_node_dict = {}
        FWidgetExtractor.get_resource(self.parent_node,
                                      image_node_dict=image_node_dict)
        # Generate
        FImageParser.generate_flutter_source(
            figma_py=self.figma_py,
            file_id=self.file_id,
            folder=folder,
            image_node_dict=image_node_dict,
            svg_maximum_size=svg_maximum_size,
            need_print=need_print
        )

    @staticmethod
    def generate_flutter_source(figma_py,
                                file_id,
                                folder: str,
                                image_node_dict: dict = None,
                                svg_maximum_size=None,
                                need_print=True):
        if image_node_dict is None:
            error_printer('Could not found any images!')
            return
        warning_printer('Generating images...')
        # Split images into SVG (small size/icon) and PNG (2x scale with larger image)
        svg_images, scaled_2x_png_images = FImageParser.detect_image_extension(
            image_node_dict=image_node_dict,
            svg_maximum_size=svg_maximum_size
        )
        # Download images
        downloaded_image_paths = []
        downloaded_image_paths.extend(download_images(
            figma=figma_py,
            file_id=file_id,
            folder=folder,
            images=svg_images,
            format_image='svg'
        ))
        downloaded_image_paths.extend(download_images(
            figma=figma_py,
            file_id=file_id,
            folder=folder,
            images=scaled_2x_png_images
        ))
        if not downloaded_image_paths:
            print('Could not found any images!')
        if need_print:
            beauty_printer(title='Successfully downloaded images file',
                           content=downloaded_image_paths)

    @staticmethod
    def detect_image_extension(image_node_dict: dict = None,
                               svg_maximum_size=None):
        """
        :param image_node_dict: Keeps all detected images/icons from a Figma node
        :param svg_maximum_size: Maximum size as we could consider this image as a svg
        :return: A tuple contains (0: all svg images, 1: all png 2x scaled images)
        """
        svg_images = {}
        scaled_2x_png_images = {}
        if image_node_dict is None:
            return svg_images, scaled_2x_png_images

        nn_svg_maximum_size = svg_maximum_size or SVG_MAXIMUM_SIZE
        for image_id, node in image_node_dict.items():
            rect_node = FWidgetExtractor.as_rect(node)
            if rect_node.width < nn_svg_maximum_size \
                    and rect_node.height < nn_svg_maximum_size:
                svg_images[image_id] = node
            else:
                scaled_2x_png_images[image_id] = node
        return svg_images, scaled_2x_png_images


# -------------------------------------------------------------------------
# FLUTTER FONT
# -------------------------------------------------------------------------
class FFontParser(FParser):
    def generate(self,
                 api_key,
                 folder: str = None,
                 basic_font=False,
                 need_print=True):
        # Fetch fonts
        font_node_list = []
        FWidgetExtractor.get_resource(self.parent_node,
                                      font_node_list=font_node_list)
        # Generate
        FFontParser.generate_flutter_source(font_node_list=font_node_list,
                                            folder=folder,
                                            api_key=api_key,
                                            basic_font=basic_font,
                                            need_print=need_print)

    @staticmethod
    def generate_flutter_source(font_node_list: list = None,
                                folder: str = None,
                                api_key: str = None,
                                basic_font=False,
                                need_print=True):
        if font_node_list is None:
            error_printer('Could not found any font styles!')
            return
        warning_printer('Generating fonts...')
        google_font_dict = TextStyle.extract_google_font_family(
            font_node_list,
            basic_font
        )
        downloaded_font_paths = download_fonts(
            google_font_dict=google_font_dict,
            api_key=api_key,
            folder=folder)
        if not downloaded_font_paths:
            print('Could not found any font styles!')
        if need_print:
            beauty_printer(title='Successfully downloaded fonts file',
                           content=downloaded_font_paths)


# -------------------------------------------------------------------------
# FLUTTER WIDGET
# -------------------------------------------------------------------------
class FWidgetParser(FParser):
    def __init__(self, parent_node, figma_py, ui_config: UiConfig):
        super().__init__(parent_node, figma_py, ui_config)
        # Strings data structure: { 'text_id' : 'LocaleKeys.camelCase' }
        self.strings_mapping: dict[str: str] = {}
        # Colors data structure: { 'hex_color': 'ColorName.camelCaseNamedColor' }
        self.colors_mapping: dict[str: str] = {}
        # Images data structure: { 'image_node_id':
        # { 'path' : 'Assets.images.svg/img.camelCase', 'format': 'svg' } }
        self.images_mapping: dict[str: NamedImage] = {}

    def generate(self,
                 need_preview=True,
                 gen_string=True,
                 gen_color=True,
                 gen_image=True,
                 gen_font=True):
        """
        Generate all resource from this [figma_node] which we need to create widgets
        in Flutter, such as images/strings/colors/fonts

        :param need_preview: Whether need to printout to console or not
        :param gen_image: Whether need to generate images or not
        :param gen_string: Whether need to generate strings or not
        :param gen_color: Whether need to generate colors or not
        :param gen_font: Whether need to generate fonts or not
        :return:
        """
        # Reset data
        self.strings_mapping.clear()
        self.colors_mapping.clear()
        self.images_mapping.clear()
        FRawWidget.SCREEN_DESIGN_WIDTH = self.figma_config.screen_width
        FRawWidget.SCREEN_DESIGN_HEIGHT = self.figma_config.screen_height
        # Detect screen node
        custom_widget_nodes = []
        FWidgetExtractor.find_all_custom_widget_node(self.parent_node,
                                                     custom_widget_nodes)
        # Fetching resources from root node
        gen_content_dict = {}
        text_dict = OrderedDict()
        hex_color_list = []
        image_node_dict = {}
        font_node_list = []
        FWidgetExtractor.get_resource(self.parent_node,
                                      text_dict=text_dict,
                                      text_in_page_dict=OrderedDict(),
                                      hex_color_list=hex_color_list,
                                      image_node_dict=image_node_dict,
                                      font_node_list=font_node_list)
        named_color_dict = fetch_named_colors(hex_color_list or [])
        # Convert resource to Flutter based mapping
        # Convert strings
        for custom_widget_name, text_in_widget in text_dict.items():
            for text_id, text_name_with_content in text_in_widget.items():
                localized_text = NamedLocaleText.localized(
                    custom_widget_name=custom_widget_name,
                    text_name=text_name_with_content[0]
                )
                if not localized_text:
                    continue
                self.strings_mapping[text_id] = localized_text
        # Convert colors
        self.colors_mapping = {
            hex_c: named_c.flutter_gen_color
            for
            hex_c, named_c in named_color_dict.items()}
        # Convert images
        svg_images, scaled_2x_png_images = FImageParser.detect_image_extension(
            image_node_dict=image_node_dict,
            svg_maximum_size=self.figma_config.maximum_svg_size_in_px
        )
        self.images_mapping = {
            image_id: NamedImage(
                image_name=image_node['name'],
                image_format='svg'
            )
            for
            image_id, image_node in
            svg_images.items() if 'name' in image_node}
        self.images_mapping.update({
            image_id: NamedImage(
                image_name=image_node['name'],
                image_format='png'
            )
            for
            image_id, image_node in
            scaled_2x_png_images.items() if 'name' in image_node})

        # Generate strings
        if gen_string:
            FStringParser.generate_flutter_source(
                strings=text_dict,
                locale_folder_path=self.assets_config.translations,
                pivot_locale=self.google_config.translate.main_locale,
                need_print=need_preview)
        # Generate colors
        if gen_color:
            FColorParser.generate_flutter_source(
                named_color_dict,
                folder=self.assets_config.colors,
                need_print=need_preview)
        # Generate images
        if gen_image:
            FImageParser.generate_flutter_source(
                self.figma_py,
                self.file_id,
                folder=self.assets_config.images,
                image_node_dict=image_node_dict,
                svg_maximum_size=self.figma_config.maximum_svg_size_in_px,
                need_print=need_preview)
        # Generate fonts
        if gen_font:
            FFontParser.generate_flutter_source(
                font_node_list=font_node_list,
                folder=self.assets_config.fonts,
                api_key=self.google_config.font.api_key,
                basic_font=self.google_config.font.basic_font,
                need_print=need_preview)
        # Generate needed resource paths before generating widgets
        if gen_string:
            bash_execute(FLUTTER_GEN_LOCALE)
        if gen_color or gen_image or gen_font:
            bash_execute(FLUTTER_GEN)

        # Fetching Flutter Raw Widgets
        for widget_node in custom_widget_nodes:
            root_view = self._make_tree_view(widget_node)
            if not root_view or not root_view.children:
                continue
            self._optimize_tree_views(root_view)
            if root_view.type == FWidgetType.PAGE:
                # Update screen design size
                FRawWidget.SCREEN_DESIGN_WIDTH = root_view.rect.width
                FRawWidget.SCREEN_DESIGN_HEIGHT = root_view.rect.height
            # Convert resources into a mapping dict for Flutter generator
            f_root_widget = FConverter.convert(root_view)
            FConverter.generate_widgets(
                current_widget=f_root_widget,
                gen_content_dict=gen_content_dict,
                strings_mapping=self.strings_mapping,
                colors_mapping=self.colors_mapping,
                images_mapping=self.images_mapping
            )
            # TODO: In case we just have tpage without page, then generate new page
            if need_preview:
                need_print_custom_widget = root_view.type != FWidgetType.PAGE \
                                           and root_view.type != FWidgetType.TAB_PAGE
                FWidgetParser.preview_tree_view(
                    root_view,
                    need_print_custom_widget=need_print_custom_widget
                )

        # Generate widgets and files
        warning_printer('Generating widgets...')
        FGenerator.generate_ui_file(gen_content_dict)

        # Format Dart code
        bash_execute(command=FLUTTER_FORMAT)

    def _make_tree_view(self, current_node: dict):
        f_widget = FWidgetExtractor.figma_to_raw_flutter(current_node)
        if not f_widget:
            return None
        # Filter useless node
        child_nodes = current_node['children'] \
            if 'children' in current_node \
            else []
        if len(child_nodes) == 0 \
                and f_widget.type == FWidgetType.IGNORED:
            # Ignore this single view has ignored type
            return None

        # Filter customized node
        if f_widget.is_custom_widget:
            current_id = current_node['id'] if 'id' in current_node else ''
            customized_node = current_node.copy()
            customized_node['id'] = f'{current_id}+'
            customized_node['name'] = ''
            first_child_view = self._make_tree_view(customized_node)
            if first_child_view:
                f_widget.children.append(first_child_view)
            return f_widget

        # Filter node that presented as a single view
        if not f_widget.type.is_view_group() \
                and f_widget.type != FWidgetType.IGNORED:
            # Will return defined Flutter single view type
            return f_widget

        # Create children
        child_views = []
        for child_node in child_nodes:
            child_view = self._make_tree_view(child_node)
            if child_view:
                child_views.append(child_view)
        if len(child_views) == 0:
            # Ignore the view which doesn't contain any FlutterView as child
            return None
        FWidgetParser.detect_view_group(f_widget, child_views)
        return f_widget

    def _optimize_tree_views(self, current_view: FRawWidget):
        """
        Merge all child views from parent view into grandpa view if
        grandpa view has the same type as parent view. Only consider flex
        view group type.

        Note: WRAP type will absorb all the other flex types

        :param current_view:
        :return:
        """
        if not current_view:
            return

        for index, child in enumerate(current_view.children):
            self._optimize_tree_views(child)
            if not child.type.is_flex_view_group():
                continue
            if current_view.type != FWidgetType.WRAP \
                    and child.type != current_view.type:
                continue
            # Merge all child views of this `child` into its `parent`
            # (i.e. current_view in this case)
            current_view.children.pop(index)
            for sub_index, sub_child in enumerate(child.children):
                current_view.children.insert(index + sub_index, sub_child)
        current_view.optimize_view_type()
        # Final sorting by left/top to get correct order of row/column
        # Because in some cases finding vertices algorithm has exceptions
        if current_view.type == FWidgetType.ROW:
            current_view.children.sort(key=lambda c: c.rect.left)
        if current_view.type == FWidgetType.COLUMN:
            current_view.children.sort(key=lambda c: c.rect.top)

    @staticmethod
    def detect_view_group(flutter_view: FRawWidget,
                          child_views: list[FRawWidget]):
        """
        Defined by consistent view group type such as ListView/GridView

        or

        Calculate position of child views to detect type of this view as a flex view group
        such as Stack/Column/Row/Wrap

        :param flutter_view: Parent view
        :param child_views: Child views which are considered to append in the [flutter_view] by order. Sorted by ascending z-axis from Figma
        :return: Parent view after detected type and grouped ordered child views
        """
        if not flutter_view:
            return None

        if not child_views or len(child_views) == 0:
            return flutter_view

        # ----Detect background of flutter_view----
        # Usually is the first child which has lowest z-value
        # Ignore the background view
        background_view = None
        clutter_views = []
        # Filter background view and the other clutter views
        for child_v in child_views:
            # Always store the latest background
            if child_v.type == FWidgetType.BACKGROUND:
                background_view = child_v
            else:
                clutter_views.append(child_v)
        if background_view:
            flutter_view.type = FWidgetType.STACK
            flutter_view.children.clear()
            flutter_view.children.append(background_view)
            group_other_views = FWidgetParser.make_group_view(flutter_view,
                                                              clutter_views)
            if group_other_views:
                FWidgetParser.detect_view_group(group_other_views,
                                                clutter_views)
                flutter_view.children.append(group_other_views)
            return flutter_view

        # Filter useless view
        if not clutter_views or len(clutter_views) == 0:
            return flutter_view

        # ----Detect Stack View----
        # clutter_views is sorted by ascending z-axis because child_views is sorted by same way
        # Check overlapping from higher z to lower z
        for higher_z_index in range(len(clutter_views) - 1, 0, -1):
            higher_view = clutter_views[higher_z_index]
            for lower_z_index in range(higher_z_index - 1, -1, -1):
                lower_view = clutter_views[lower_z_index]
                if higher_view.rect.is_contained_by(lower_view.rect) or \
                        higher_view.rect.is_intersected_with(
                            lower_view.rect):
                    flutter_view.type = FWidgetType.STACK
                    flutter_view.children.clear()
                    lower_clutter_views = clutter_views[:higher_z_index]
                    group_other_views = FWidgetParser.make_group_view(
                        flutter_view,
                        lower_clutter_views
                    )
                    if group_other_views:
                        FWidgetParser.detect_view_group(group_other_views,
                                                        lower_clutter_views)
                        flutter_view.children.append(group_other_views)
                    flutter_view.children.append(higher_view)
                    return flutter_view

        # ----Detect consistent view group----
        if flutter_view.type.is_consistent_view_group():
            flutter_view.children = [clutter_views[0]]
            return flutter_view

        # ----Default is CONTAINER if child_views contains only 1 child----
        if len(clutter_views) == 1:
            if flutter_view.type == FWidgetType.IGNORED:
                flutter_view.type = FWidgetType.CONTAINER
                flutter_view.children = clutter_views
            return flutter_view

        # ----Detect Column/Row/Wrap View----
        # Step 1: Find 4 vertices of this view
        #       -Y
        #        ↑
        # -X ← O(x, y) → X
        #        ↓
        #        Y
        top_left_most, bottom_left_most, top_right_most, bottom_right_most = None, None, None, None
        cur_tl, cur_bl, cur_tr, cur_br = None, None, None, None
        for child_v in clutter_views:
            v_rect = child_v.rect
            # Find the top left most: min(x+y)
            tl = v_rect.left + v_rect.top
            if not cur_tl or tl < cur_tl:
                top_left_most = child_v
                cur_tl = tl

            # Find the bottom left most: min(x-y)
            bl = v_rect.left - v_rect.bottom
            if not cur_bl or bl < cur_bl:
                bottom_left_most = child_v
                cur_bl = bl

            # Find the top right most: max(x-y)
            tr = v_rect.right - v_rect.top
            if not cur_tr or tr > cur_tr:
                top_right_most = child_v
                cur_tr = tr

            # Find the bottom right most: max(x+y)
            br = v_rect.right + v_rect.bottom
            if not cur_br or br > cur_br:
                bottom_right_most = child_v
                cur_br = br

        fill_left_side = top_left_most == bottom_left_most
        fill_top_side = top_left_most == top_right_most
        fill_right_side = top_right_most == bottom_right_most
        fill_bottom_side = bottom_left_most == bottom_right_most

        # Step 2: Find side-positional view which occupy 2 vertices
        def _find_ordered_views():
            # WRAP type just need to take the first child
            if flutter_view.type is FWidgetType.WRAP:
                return [top_left_most]

            # If each view occupies just 1 vertex then flutter_view is a COLUMN by default
            if not fill_left_side and not fill_top_side \
                    and not fill_right_side and not fill_bottom_side:
                # Find highest top vertex
                flutter_view.type = FWidgetType.COLUMN
                return [
                    top_left_most
                    if top_left_most.rect.top > top_right_most.rect.top
                    else top_right_most
                ]

            # If 1 view occupies 4 vertices e.g. 1 view has width height too large
            # than the other views, then compare relative position with the other views
            if fill_left_side and fill_top_side \
                    and fill_right_side and fill_bottom_side:
                for clutter_v in [v for v in clutter_views if
                                  v != top_left_most]:
                    if clutter_v.rect.right >= top_left_most.rect.right:
                        flutter_view.type = FWidgetType.ROW
                        return [top_left_most]
                    if clutter_v.rect.bottom >= top_left_most.rect.bottom:
                        flutter_view.type = FWidgetType.COLUMN
                        return [top_left_most]

            # Check whether 1 view occupies 3 vertices then we need to compare relative
            # position with the other view which occupies remaining vertex
            if fill_left_side and fill_top_side:
                # Compare relative position with bottom right most to detect row or column
                flutter_view.type = FWidgetType.ROW \
                    if bottom_right_most.rect.left >= top_left_most.rect.right \
                    else FWidgetType.COLUMN
                return [top_left_most]

            if fill_left_side and fill_bottom_side:
                flutter_view.type = FWidgetType.ROW \
                    if top_right_most.rect.left >= top_left_most.rect.right \
                    else FWidgetType.COLUMN
                return [top_left_most]

            if fill_right_side and fill_top_side:
                flutter_view.type = FWidgetType.ROW \
                    if bottom_left_most.rect.right <= top_right_most.rect.left \
                    else FWidgetType.COLUMN
                return [top_left_most]

            if fill_right_side and fill_bottom_side:
                flutter_view.type = FWidgetType.ROW \
                    if top_left_most.rect.right <= top_right_most.rect.left \
                    else FWidgetType.COLUMN
                return [top_left_most]

            # Otherwise in normal case, when 1 view occupies 2 vertices then
            # we could define the view group's type
            ordered_views = []
            if fill_left_side:
                flutter_view.type = FWidgetType.ROW
                ordered_views.append(top_left_most)
            if fill_top_side:
                flutter_view.type = FWidgetType.COLUMN
                ordered_views.append(top_left_most)
            if fill_right_side:
                flutter_view.type = FWidgetType.ROW
                ordered_views.append(bottom_right_most)
            if fill_bottom_side:
                flutter_view.type = FWidgetType.COLUMN
                ordered_views.append(bottom_right_most)
            return ordered_views

        # Find and disabled redundant sides
        ordered_child_views = _find_ordered_views()
        if len(ordered_child_views) == 1:
            if ordered_child_views[0] == top_left_most:
                if flutter_view.type == FWidgetType.ROW:
                    fill_right_side = False
                if flutter_view.type == FWidgetType.COLUMN:
                    fill_bottom_side = False
            if ordered_child_views[0] == bottom_right_most:
                if flutter_view.type == FWidgetType.ROW:
                    fill_left_side = False
                if flutter_view.type == FWidgetType.COLUMN:
                    fill_top_side = False

        # Step 3: Group the other views into a new frame then recur to (step 1)
        un_ordered_child_views = [child for child in clutter_views if
                                  child not in ordered_child_views]

        # Final check to filter some case that finding vertices algorithm
        # could not handle e.g view has width to long so it will occupy
        # left side but actually it just occupy top left vertex in design file
        removed_ordered_child_views = []
        for un_ordered_child in un_ordered_child_views:
            if flutter_view.type == FWidgetType.ROW:
                if fill_left_side \
                        and un_ordered_child.rect.left < top_left_most.rect.right:
                    if top_left_most in ordered_child_views:
                        ordered_child_views.remove(top_left_most)
                    removed_ordered_child_views.append(top_left_most)
                if fill_right_side \
                        and un_ordered_child.rect.right > bottom_right_most.rect.left:
                    if bottom_right_most in ordered_child_views:
                        ordered_child_views.remove(bottom_right_most)
                    removed_ordered_child_views.append(bottom_right_most)
            if flutter_view.type == FWidgetType.COLUMN:
                if fill_top_side \
                        and un_ordered_child.rect.top < top_left_most.rect.bottom:
                    if top_left_most in ordered_child_views:
                        ordered_child_views.remove(top_left_most)
                    removed_ordered_child_views.append(top_left_most)
                if fill_bottom_side \
                        and un_ordered_child.rect.bottom > bottom_right_most.rect.top:
                    if bottom_right_most in ordered_child_views:
                        ordered_child_views.remove(bottom_right_most)
                    removed_ordered_child_views.append(bottom_right_most)
        un_ordered_child_views.extend(removed_ordered_child_views)

        # Filter useless view group
        if len(ordered_child_views) == 0:
            return flutter_view

        # Make a clutter view group from unordered views
        group_other_views = FWidgetParser.make_group_view(flutter_view,
                                                          un_ordered_child_views)

        # Step 4: Detect ordered children of view group of this view
        if group_other_views:
            FWidgetParser.detect_view_group(group_other_views,
                                            un_ordered_child_views)
            if flutter_view.type == FWidgetType.WRAP:
                ordered_child_views.append(group_other_views)
            elif flutter_view.type == FWidgetType.COLUMN:
                ordered_child_views.insert(
                    1 if len(ordered_child_views) > 0 and ordered_child_views[
                        0] == top_left_most else 0,
                    group_other_views
                )
            elif flutter_view.type == FWidgetType.ROW:
                ordered_child_views.insert(
                    1 if len(ordered_child_views) > 0 and ordered_child_views[
                        0] == top_left_most else 0,
                    group_other_views
                )
        flutter_view.children = ordered_child_views
        return flutter_view

    @staticmethod
    def make_group_view(parent_view: FRawWidget,
                        other_child_views: [FRawWidget]):
        total_other_child = len(other_child_views)
        if total_other_child == 0:
            return None
        first_child = other_child_views[0]
        if total_other_child == 1:
            return first_child
        left_most = first_child.rect.left
        top_most = first_child.rect.top
        right_most = first_child.rect.right
        bottom_most = first_child.rect.bottom
        for index_other_child in range(total_other_child):
            other_child = other_child_views[index_other_child]
            if left_most > other_child.rect.left:
                left_most = other_child.rect.left
            if top_most > other_child.rect.top:
                top_most = other_child.rect.top
            if right_most < other_child.rect.right:
                right_most = other_child.rect.right
            if bottom_most < other_child.rect.bottom:
                bottom_most = other_child.rect.bottom
        v_type = FWidgetType.WRAP \
            if parent_view.type is FWidgetType.WRAP \
            else FWidgetType.IGNORED
        return FRawWidget(
            figma_node={
                'id': f'{parent_view.id_view}+',
                'name': '',
                'absoluteBoundingBox': {
                    'x': left_most,
                    'y': top_most,
                    'width': right_most - left_most,
                    'height': bottom_most - top_most
                },
                'children': [child_dict.figma_node for child_dict in
                             other_child_views]
            },
            v_type=v_type,
            children=other_child_views
        )

    @staticmethod
    def preview_tree_view(root_view: FRawWidget,
                          need_print_custom_widget=True):

        def print_view(current_view: FRawWidget,
                       content_printer: [str],
                       level=0,
                       level_still_has_child: list[int] = None,
                       is_last_child=False,
                       defined_prefix=None):
            level_template = '│' + 5 * space_char
            space_template = 6 * space_char
            box_char = '└' if is_last_child else '├'
            prefix = '' if not defined_prefix \
                else f'{defined_prefix}{box_char}───'

            # Find next level prefix
            def _this_level_still_has_child(cur_level):
                for lv in level_still_has_child:
                    if cur_level == lv:
                        return True
                return False

            prefix_list = []
            for i in range(level):
                prefix_list.append(level_template
                                   if _this_level_still_has_child(i)
                                   else space_template)
            next_prefix = space_template + ''.join(prefix_list)
            hide_custom_type = current_view.type == FWidgetType.NEW_WIDGET \
                               and not need_print_custom_widget
            # Append child line
            child_line = f'{fixed_space_line}{prefix}{current_view.present}'
            content_printer.append(child_line)
            # Append space line
            child_views = current_view.children
            total_children = len(child_views)
            space_line_template = (defined_prefix or '') \
                if is_last_child and total_children == 0 else next_prefix
            space_line_last_char = '' if total_children == 0 \
                                         or hide_custom_type else '│'
            space_line = f'{fixed_space_line}{space_line_template}{space_line_last_char}'
            content_printer.append(space_line)
            # Append children view line
            if hide_custom_type:
                return
            for index, child in enumerate(child_views):
                is_last_child = index == total_children - 1
                if is_last_child and _this_level_still_has_child(level):
                    level_still_has_child.remove(level)
                elif not is_last_child \
                        and not _this_level_still_has_child(level):
                    level_still_has_child.append(level)
                print_view(
                    child,
                    content_printer,
                    level + 1,
                    level_still_has_child=level_still_has_child,
                    is_last_child=is_last_child,
                    defined_prefix=next_prefix
                )

        if not root_view:
            return
        # Show preview of generated view in console
        current_level = 0
        space_char = ' '
        fixed_space_line = 8 * space_char
        content = ['\n']
        print_view(root_view,
                   content,
                   current_level,
                   level_still_has_child=[])
        content.append('\n')
        beauty_printer(
            title=f'Preview for {root_view.full_name!r} widget',
            content=content,
            need_line_separate=False,
            max_char_inline=100
        )



class DefinedErrorMessage:
    ERROR_SUPPORT_TEMPLATE = 'Currently we are just supporting ' \
                             '[bloc/model/network/ui] template!'
    ERROR_SUPPORT_SOURCE_INPUT = 'Currently we are just supporting ' \
                                 '[pasteboard] and [file path]!'
    ERROR_SUPPORT_FORMAT = 'Currently we are just supporting ' \
                           '[json] and [yaml] format!'
    ERROR_SUPPORT_COMPONENT = 'Currently we are just supporting ' \
                              '[image/string/color/widget] component!'
    FILE_NOT_FOUND = '❌  File not found ❌ '

    @staticmethod
    def print_error(e: any):
        print(f'❌  Occurs an error - {str(e)} ❌ ')



from setuptools import setup, find_packages

from kntgen.__main__ import __version__

with open('README.md', 'r') as readme:
    long_description = readme.read()

setup(
    name='kntgen',
    version=__version__,
    author='kiendtvt1411',
    author_email='dtvtdevelopersk58@gmail.com',
    description='A generator for Dart code using by Appixi\'s projects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/appixi/knt_dart_generator',
    packages=find_packages(),
    entry_points='''
        [console_scripts]
        kntgen=kntgen.__main__:main
    ''',
    install_requires=[
        'arghandler>=1.2',
        'regex',
        'googletrans==4.0.0rc1',
        'requests',
        'pyyaml',
        'tqdm',
        'jinja2>=2.10',
        'pydantic',
        'sheetfu'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    include_package_data=True
)



# coding=utf-8

import subprocess


def pasteboard_read():
    return subprocess.check_output('pbpaste', env={'LANG': 'en_US.UTF-8'}) \
        .decode('utf-8')


def pasteboard_write(output):
    process = subprocess.Popen(
        'pbcopy',
        env={'LANG': 'en_US.UTF-8'},
        stdin=subprocess.PIPE
    )
    process.communicate(output.encode('utf-8'))
